'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var axios = require('axios');
var sharp = require('sharp');
var path = require('path');
var fs = require('fs');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var axios__default = /*#__PURE__*/_interopDefaultLegacy(axios);
var sharp__default = /*#__PURE__*/_interopDefaultLegacy(sharp);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);

/**
 * Validate all the parameters passed in to mapboxed.
 *
 * Returns true if all parameters are valid, throws an Error if any parameter is invalid.
 *
 * Validation rules:
 *  1  - Mapbox token must be valid
 *  2  - format must be one of: jpg90, jpg80, jpg70, png256, png128, png64, png32, png
 *  3  - zoom must be an integer larger than 0
 *  4A - If location is set, then we reach for mapbox's API to get the boundingbox for
 *       coordinate values.
 *  4B - x values must be between -180 and 180, y values between -90 and 90.
 *       x2 must be greater than x1, y2 must be less than y1.
 *  5  - tileset must be set
 *  6  - parallel must be an integer greater than 0
 *
 * @param params
 */
const validateParams = async (params) => {
  if (typeof params === 'undefined') throw new Error('params is undefined');

  // Read params
  const { tileset, key, format, zoom, res2x, parallel, location } = params;
  let { X1, X2, Y1, Y2 } = params;

  // //////////////////////////////////////////////////////
  // 1. Mapbox token must be valid.
  const MAPBOX_TOKEN = key || (typeof process === 'object' && process.env.MAPBOX_TOKEN);
  if (!MAPBOX_TOKEN) {
    throw new Error(
      'No mapbox token found, please pass your mapbox API token via --key or set the MAPBOX_TOKEN environment variable.',
    );
  }

  // //////////////////////////////////////////////////////
  // 2. format must be one of: jpg90, jpg80, jpg70, png256, png128, png64, png32, png
  const validFormats = ['jpg90', 'jpg80', 'jpg70', 'png256', 'png128', 'png64', 'png32', 'png'];
  if (validFormats.indexOf(format) < 0) {
    throw new Error(`format must be one of (${validFormats.join(' | ')})`);
  }

  // //////////////////////////////////////////////////////
  // 3. zoom must be an integer larger than 0
  const zoomInt = parseInt(zoom, 10);
  if (zoomInt < 0 || Number.isNaN(zoomInt)) throw new Error('zoom must be an integer greater than or equal to 0');
  const n = 2 ** zoomInt;

  // //////////////////////////////////////////////////////
  // 4. Get the location information

  // 4A. If the location flag is specified, we query mapbox's API for
  //     geocoded coordinates.
  if (location) {
    // Download tile using axios
    const response = await axios__default['default']({
      url: `https://api.mapbox.com/geocoding/v5/mapbox.places/${location}.json?access_token=${MAPBOX_TOKEN}`,
      method: 'GET',
    });

    // Get the bbox from the first feature.
    if (response.data && response.data.features && response.data.features[0]) {
      [X1, Y2, X2, Y1] = response.data.features[0].bbox;
    }
  }

  // 4B. x values must be between -180 and 180, y values between -90 and 90.
  //     x2 must be greater than x1, y2 must be less than y1.
  const bounds = [
    parseFloat(X1), parseFloat(Y1), parseFloat(X2), parseFloat(Y2),
  ];

  if (
    bounds[0] >= bounds[2] || bounds[1] <= bounds[3]
      || bounds[0] > 180 || bounds[0] < -180 || bounds[2] > 180 || bounds[2] < -180
      || bounds[1] > 90 || bounds[1] < -90 || bounds[3] > 90 || bounds[3] < -90
  ) {
    throw new Error(
      `invalid latitude / longitude values. x values must be between -180 and 180, y values between -90 and 90. 
      x2 must be greater than x1, y2 must be less than y1.`,
    );
  }

  // eslint-disable-next-line no-restricted-globals
  if (bounds.some((c) => isNaN(c))) throw new Error('Coordinates could not be found / used.');

  // //////////////////////////////////////////////////////
  // 5. tileset must be set
  if (!tileset) throw new Error('tileset must be set');

  // //////////////////////////////////////////////////////
  // 6. parallel must be set
  const parallelInt = parseInt(parallel, 10);
  if (parallelInt <= 0 || Number.isNaN(parallelInt)) throw new Error('parallel must be an integer greater than 0');

  // //////////////////////////////////////////////////////
  // Return processed values
  return {
    MAPBOX_TOKEN, bounds, zoom: zoomInt, n, tileset, res2x: !!res2x, parallel: parallelInt, format,
  };
};

/**
 * Slippy Tileset calculation functions
 *
 * https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Lon..2Flat._to_tile_numbers
 */
// Convert lat / long to tile location
const deg2num = (lonDeg, latDeg, n) => {
  const latRad = latDeg * (Math.PI / 180);
  const xtile = Math.floor(n * ((lonDeg + 180) / 360));
  const ytile = Math.floor(n * ((1 - (Math.asinh(Math.tan(latRad))) / Math.PI) / 2));
  return [xtile, ytile];
};

// Convert tile location to lat / long
const num2deg = (x, y, n) => {
  const lonDeg = (x / n) * 360 - 180;
  const latRad = Math.atan(Math.sinh(Math.PI * (1 - (2 * y) / n)));
  const latDeg = latRad * (180 / Math.PI);
  return [lonDeg, latDeg];
};

/**
 * Generate the tiles we need for this instance
 */
const generateTiles = ({ bounds, n }) => {
  let tiles = [];
  const min = deg2num(bounds[0], bounds[1], n);
  const max = deg2num(bounds[2], bounds[3], n);
  const xRange = [min[0], max[0]].sort();
  const yRange = [min[1], max[1]].sort();
  const xTilesLength = xRange[1] - xRange[0];
  const yTilesLength = yRange[1] - yRange[0];
  for (let x = xRange[0]; x <= xRange[1]; x += 1) {
    for (let y = yRange[0]; y <= yRange[1]; y += 1) {
      tiles = tiles.concat([[x, y]]);
    }
  }
  return { tiles, min, max, xTilesLength, yTilesLength };
};

/**
 * Mapbox Tile URL
 */
const tileURL = ({ tileset, zoom, res2x, format, MAPBOX_TOKEN }, x, y) => `https://api.mapbox.com/v4/${tileset}/${zoom}/${x}/${y}${res2x ? '@2x' : ''}.${format}?access_token=${MAPBOX_TOKEN}`;

/**
 * Server
 *
 * Functions specific to downloading / assembling tiles on the server or through the command line.
 */

/**
 * The path to the individual tile image on the filesystem.
 *
 * @param tileset
 * @param zoom
 * @param res2x
 * @param format
 * @param x
 * @param y
 * @returns {string}
 */
const imgPath = ({ tileset, zoom, res2x, format }, x, y) => `./.mapboxed/${zoom}/${tileset}-${x}-${y}${res2x ? '@2x' : ''}.${format.replace(/[0-9]/g, '')}`;

/**
 * Download this particular tile.
 *
 * @param params
 * @param x
 * @param y
 * @param onDownloadProgress
 * @returns {Promise<unknown>}
 */
const downloadTile = async (params, x, y, onDownloadProgress = () => {}) => {
  // Get parameters
  const { zoom, tileset, res2x, format } = params;

  // Get the tile mapbox api url
  const url = tileURL(params, x, y);

  // get and generate directory structure if it doesn't exist
  const imagesDir = path__default['default'].resolve('./.mapboxed');
  if (!fs__default['default'].existsSync(imagesDir)) fs__default['default'].mkdirSync(imagesDir);
  const zoomDir = path__default['default'].resolve(imagesDir, `${zoom}`);
  if (!fs__default['default'].existsSync(zoomDir)) fs__default['default'].mkdirSync(zoomDir);
  const tilePath = path__default['default'].resolve(
    zoomDir,
    `${tileset}-${x}-${y}${res2x ? '@2x' : ''}.${format.replace(/[0-9]/g, '')}`,
  );

  // Download write stream
  const writer = fs__default['default'].createWriteStream(tilePath);

  // Download tile using axios
  const response = await axios__default['default']({ url, method: 'GET', responseType: 'stream', onDownloadProgress });

  // Call onDownloadProgress to notify that we finished downloading
  const responseSize = parseInt(response.headers['content-length'], 10);
  onDownloadProgress({ total: responseSize, loaded: responseSize });

  // save to file
  response.data.pipe(writer);

  return new Promise((resolve, reject) => {
    writer.on('finish', resolve);
    writer.on('error', reject);
  });
};

/**
 * Takes a list of tiles and composites them into a full sharp image.
 *
 * @param params
 * @param tiles
 * @param xTilesLength
 * @param yTilesLength
 * @param cb
 * @returns {Promise<void>}
 */
const generateImage = async (params, tiles, xTilesLength, yTilesLength, cb = (d) => d) => {
  // Read the params we need
  const { format, res2x } = params;
  const tileWidth = (res2x ? 512 : 256);
  const compositeStep = 8;

  // Generate format options for sharp
  // https://sharp.pixelplumbing.com/api-output
  const formatOptions = {};
  switch (format.replace(/[0-9]/g, '')) {
    case 'jpg':
      formatOptions.quality = parseInt(format.replace(/[a-z]/g, ''), 10);
      break;
    case 'png':
      formatOptions.colors = parseInt(format.replace(/[a-z]/g, '') || '256', 10);
      break;
  }

  // base coordinates
  const [baseX, baseY] = tiles[0];

  // Create sharp object with first tile.
  let composited = sharp__default['default'](imgPath(params, baseX, baseY))
    // We want to extend the size ahead of time to the final size
    .extend({
      top: 0,
      left: 0,
      bottom: yTilesLength * tileWidth,
      right: xTilesLength * tileWidth,
      background: { r: 0, g: 0, b: 0, alpha: 0 },
    });

  // composite the rest of the tiles.
  const restTiles = tiles.slice(1);

  // Janky nested raw().toBuffer() loop in chunks of compositeStep because of
  // potential out of memory error if we do them all in the same step.
  // https://github.com/lovell/sharp/issues/2286
  await composited.raw().toBuffer(async (err, buffer, info) => {
    for (let i = 0; i < restTiles.length; i += compositeStep) {
      // eslint-disable-next-line no-await-in-loop
      const innerBuffer = await composited.raw().toBuffer();

      // Set composited to image with these tiles composited.
      composited = sharp__default['default'](innerBuffer, { raw: info })
        .composite(
          // Map compositeStep number of tiles at a time
          restTiles.slice(i, i + compositeStep)
            .map((n) => ({
              input: imgPath(params, ...n),
              left: (n[0] - baseX) * tileWidth,
              top: (n[1] - baseY) * tileWidth,
            })),
        );
    }

    // Set quality
    await cb(composited.toFormat(format.replace(/[0-9]/g, ''), formatOptions));
  });
};

exports.deg2num = deg2num;
exports.downloadTile = downloadTile;
exports.generateImage = generateImage;
exports.generateTiles = generateTiles;
exports.imgPath = imgPath;
exports.num2deg = num2deg;
exports.tileURL = tileURL;
exports.validateParams = validateParams;
